==> gotosocial-debug-0.20.3-1-x86_64.pkg.tar.zst: usr/src/debug/gotosocial/gotosocial/internal/media/processingemoji.go <==
// GoToSocial
// Copyright (C) GoToSocial Authors admin@gotosocial.org
// SPDX-License-Identifier: AGPL-3.0-or-later
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package media

import (
	"context"
	"os"

	"code.superseriousbusiness.org/gotosocial/internal/gtserror"
	"code.superseriousbusiness.org/gotosocial/internal/gtsmodel"
	"code.superseriousbusiness.org/gotosocial/internal/log"
	"code.superseriousbusiness.org/gotosocial/internal/storage"
	"code.superseriousbusiness.org/gotosocial/internal/uris"
	"code.superseriousbusiness.org/gotosocial/internal/util"
	errorsv2 "codeberg.org/gruf/go-errors/v2"
	"codeberg.org/gruf/go-runners"
)

// ProcessingEmoji represents an emoji currently processing. It exposes
// various functions for retrieving data from the process.
type ProcessingEmoji struct {
	emoji     *gtsmodel.Emoji   // processing emoji details
	instAccID string            // instance account ID
	newPathID string            // new emoji path ID to use when being refreshed
	dataFn    DataFunc          // load-data function, returns media stream
	done      bool              // done is set when process finishes with non ctx canceled type error
	proc      runners.Processor // proc helps synchronize only a singular running processing instance
	err       error             // error stores permanent error value when done
	mgr       *Manager          // mgr instance (access to db / storage)
}

// LoadEmoji blocks until the static and fullsize image has been processed, and then returns the completed emoji.
func (p *ProcessingEmoji) Load(ctx context.Context) (*gtsmodel.Emoji, error) {
	emoji, done, err := p.load(ctx)
	if !done {
		// On a context-canceled error (marked as !done), requeue for loading.
		p.mgr.state.Workers.Dereference.Queue.Push(func(ctx context.Context) {
			if _, _, err := p.load(ctx); err != nil {
				log.Errorf(ctx, "error loading emoji: %v", err)
			}
		})
	}
	return emoji, err
}

func (p *ProcessingEmoji) LoadAsync(deferred func()) *gtsmodel.Emoji {
	p.mgr.state.Workers.Dereference.Queue.Push(func(ctx context.Context) {
		if deferred != nil {
			defer deferred()
		}

		if _, _, err := p.load(ctx); err != nil {
			log.Errorf(ctx, "error loading emoji: %v", err)
		}
	})

	var pathID string
	if p.newPathID != "" {
		pathID = p.newPathID
	} else {
		pathID = p.emoji.ID
	}

	// Placeholder returns a copy of internally stored processing placeholder,
	// returning only the fields that may be known *before* completion,
	// and as such all fields which are safe to concurrently read.
	placeholder := new(gtsmodel.Emoji)
	placeholder.ID = p.emoji.ID
	placeholder.Shortcode = p.emoji.Shortcode
	placeholder.Domain = p.emoji.Domain
	placeholder.Cached = new(bool)
	placeholder.ImageRemoteURL = p.emoji.ImageRemoteURL
	placeholder.ImageStaticRemoteURL = p.emoji.ImageStaticRemoteURL
	placeholder.Disabled = p.emoji.Disabled
	placeholder.VisibleInPicker = p.emoji.VisibleInPicker
	placeholder.CategoryID = p.emoji.CategoryID

	// We specifically set placeholder path values that allow an API user to fetch the appropriate
	// emoji, even if we don't know what filetype it is yet. (since we just parse the IDs from URL path).
	//
	// This way the API caller can (in the worst case that it hasn't loaded yet) attempt to fetch the emoji,
	// then block on ProcessingEmoji{}.Load() for the processing entry it gets from a call to the dereferencer.
	placeholder.ImageURL = uris.URIForAttachment(p.instAccID, string(TypeEmoji), string(SizeOriginal), pathID, "loading")
	placeholder.ImageStaticURL = uris.URIForAttachment(p.instAccID, string(TypeEmoji), string(SizeStatic), pathID, "png")

	return placeholder
}

// load is the package private form of load() that is wrapped to catch context canceled.
func (p *ProcessingEmoji) load(ctx context.Context) (
	emoji *gtsmodel.Emoji,
	done bool,
	err error,
) {
	err = p.proc.Process(func() error {
		if done = p.done; done {
			// Already proc'd.
			return p.err
		}

		defer func() {
			// This is only done when ctx NOT cancelled.
			if done = (err == nil || !errorsv2.IsV2(err,
				context.Canceled,
				context.DeadlineExceeded,
			)); done {
				// Processing finished,
				// whether error or not!

				// Anything from here, we
				// need to ensure happens
				// (i.e. no ctx canceled).
				ctx = context.WithoutCancel(ctx)

				// On error, clean
				// downloaded files.
				if err != nil {
					p.cleanup(ctx)
				}

				// Update with latest details, whatever happened.
				e := p.mgr.state.DB.UpdateEmoji(ctx, p.emoji)
				if e != nil {
					log.Errorf(ctx, "error updating emoji in db: %v", e)
				}

				// Store values.
				p.done = true
				p.err = err
			}
		}()

		// Attempt to store media and calculate
		// full-size media attachment details.
		//
		// This will update p.emoji as it goes.
		err = p.store(ctx)
		return err
	})

	// Return a copy of emoji.
	emoji = new(gtsmodel.Emoji)
	*emoji = *p.emoji
	return
}

// store calls the data function attached to p if it hasn't been called yet,
// and updates the underlying attachment fields as necessary. It will then stream
// bytes from p's reader directly into storage so that it can be retrieved later.
func (p *ProcessingEmoji) store(ctx context.Context) error {
	// Load media from data func.
	rc, err := p.dataFn(ctx)
	if err != nil {
		return gtserror.Newf("error executing data function: %w", err)
	}

	var (
		// predfine temporary media
		// file path variables so we
		// can remove them on error.
		temppath   string
		staticpath string
	)

	defer func() {
		if err := remove(temppath, staticpath); err != nil {
			log.Errorf(ctx, "error(s) cleaning up files: %v", err)
		}
	}()

	// Drain reader to tmp file
	// (this reader handles close).
	temppath, err = drainToTmp(rc)
	if err != nil {
		return gtserror.Newf("error draining data to tmp: %w", err)
	}

	// Pass input file through ffprobe to
	// parse further metadata information.
	result, err := probe(ctx, temppath)
	if err != nil && !isUnsupportedTypeErr(err) {
		return gtserror.Newf("ffprobe error: %w", err)
	} else if result == nil {
		log.Warnf(ctx, "unsupported data type by ffprobe: %v", err)
		return nil
	}

	var ext string
	var fileType gtsmodel.FileType

	// Get abstract file type, mimetype and ext from ffprobe data.
	fileType, p.emoji.ImageContentType, ext = result.GetFileType()
	if fileType != gtsmodel.FileTypeImage {
		return gtserror.Newf("unsupported emoji filetype: %s (%s)", fileType, ext)
	}

	// Add file extension to path.
	newpath := temppath + "." + ext

	// Before ffmpeg processing, rename to set file ext.
	if err := os.Rename(temppath, newpath); err != nil {
		return gtserror.Newf("error renaming to %s - >%s: %w", temppath, newpath, err)
	}

	// Update path var
	// AFTER successful.
	temppath = newpath

	// Generate a static image from input emoji path.
	staticpath, err = ffmpegGenerateStatic(ctx, temppath)
	if err != nil {
		return gtserror.Newf("error generating emoji static: %w", err)
	}

	var pathID string
	if p.newPathID != "" {
		// This is a refreshed emoji with a new
		// path ID that this will be stored under.
		pathID = p.newPathID
	} else {
		// This is a new emoji, simply use provided ID.
		pathID = p.emoji.ID
	}

	// Calculate final emoji media file path.
	p.emoji.ImagePath = uris.StoragePathForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeOriginal),
		pathID,
		ext,
	)

	// Calculate final emoji static media file path.
	p.emoji.ImageStaticPath = uris.StoragePathForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeStatic),
		pathID,
		"png",
	)

	// Set the known emoji static content type.
	p.emoji.ImageStaticContentType = "image/png"

	// Copy temporary file into storage at path.
	filesz, err := p.mgr.state.Storage.PutFile(ctx,
		p.emoji.ImagePath,
		temppath,
		p.emoji.ImageContentType,
	)
	if err != nil {
		return gtserror.Newf("error writing emoji to storage: %w", err)
	}

	// Copy static emoji file into storage at path.
	staticsz, err := p.mgr.state.Storage.PutFile(ctx,
		p.emoji.ImageStaticPath,
		staticpath,
		p.emoji.ImageStaticContentType,
	)
	if err != nil {
		return gtserror.Newf("error writing static to storage: %w", err)
	}

	// Set final determined file sizes.
	p.emoji.ImageFileSize = int(filesz)
	p.emoji.ImageStaticFileSize = int(staticsz)

	// Generate an emoji media static URL.
	p.emoji.ImageURL = uris.URIForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeOriginal),
		pathID,
		ext,
	)

	// Generate an emoji image static URL.
	p.emoji.ImageStaticURL = uris.URIForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeStatic),
		pathID,
		"png",
	)

	// We can now consider this cached.
	p.emoji.Cached = util.Ptr(true)

	return nil
}

// cleanup will remove any traces of processing emoji from storage,
// and perform any other necessary cleanup steps after failure.
func (p *ProcessingEmoji) cleanup(ctx context.Context) {
	log.Debugf(ctx, "running cleanup of emoji %s", p.emoji.ID)

	if p.emoji.ImagePath != "" {
		// Ensure emoji file at path is deleted from storage.
		err := p.mgr.state.Storage.Delete(ctx, p.emoji.ImagePath)
		if err != nil && !storage.IsNotFound(err) {
			log.Errorf(ctx, "error deleting %s: %v", p.emoji.ImagePath, err)
		}
	}

	if p.emoji.ImageStaticPath != "" {
		// Ensure emoji static file at path is deleted from storage.
		err := p.mgr.state.Storage.Delete(ctx, p.emoji.ImageStaticPath)
		if err != nil && !storage.IsNotFound(err) {
			log.Errorf(ctx, "error deleting %s: %v", p.emoji.ImageStaticPath, err)
		}
	}

	// Unset processor-calculated fields.
	p.emoji.ImageStaticContentType = ""
	p.emoji.ImageStaticFileSize = 0
	p.emoji.ImageStaticPath = ""
	p.emoji.ImageStaticURL = ""
	p.emoji.ImageContentType = ""
	p.emoji.ImageFileSize = 0
	p.emoji.ImagePath = ""
	p.emoji.ImageURL = ""

	// Ensure marked as not cached.
	p.emoji.Cached = util.Ptr(false)
}

==> gotosocial-debug-0.21.0-1-x86_64.pkg.tar.zst: usr/src/debug/gotosocial/gotosocial/internal/media/processingemoji.go <==
// GoToSocial
// Copyright (C) GoToSocial Authors admin@gotosocial.org
// SPDX-License-Identifier: AGPL-3.0-or-later
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package media

import (
	"context"
	"os"

	"code.superseriousbusiness.org/gopkg/log"
	"code.superseriousbusiness.org/gotosocial/internal/gtserror"
	"code.superseriousbusiness.org/gotosocial/internal/gtsmodel"
	"code.superseriousbusiness.org/gotosocial/internal/storage"
	"code.superseriousbusiness.org/gotosocial/internal/uris"
	"codeberg.org/gruf/go-errors/v2"
	errorsv2 "codeberg.org/gruf/go-errors/v2"
	"codeberg.org/gruf/go-runners"
)

// ProcessingEmoji represents an emoji currently processing. It exposes
// various functions for retrieving data from the process.
type ProcessingEmoji struct {

	// processing emoji details.
	emoji *gtsmodel.Emoji

	// instance account ID, used to
	// construct emoji storage path.
	instAccID string

	// new emoji path ID to
	// use when being refreshed.
	newPathID string

	// load data function,
	// returns media stream.
	dataFn DataFunc

	// proc helps synchronize only a
	// singular running processing instance
	proc runners.Processor

	// error stores permanent value when done,
	// or alternatively may store a preset stubError{}
	// value with details to allow skipping processing.
	err error

	// mgr instance, for access to
	// db / storage during processing
	mgr *Manager

	// done is set when process finishes
	// with non ctx canceled type error
	done bool
}

// Load blocks until the static and fullsize image has been processed, and then returns the completed emoji.
func (p *ProcessingEmoji) Load(ctx context.Context) (*gtsmodel.Emoji, error) {
	emoji, _, err := p.load(ctx)
	return emoji, err
}

// MustLoad blocks until the static and fullsize image has been processed, and then returns the completed
// emoji. On context cancelled this will enqueue the load for asynchronous operation by dereference worker.
func (p *ProcessingEmoji) MustLoad(ctx context.Context) (*gtsmodel.Emoji, error) {
	emoji, done, err := p.load(ctx)
	if !done {
		// On a context-canceled error (marked as !done), requeue for loading.
		p.mgr.state.Workers.Dereference.Queue.Push(func(ctx context.Context) {
			log.Warnf(ctx, "reprocessing emoji %s after canceled ctx", p.emoji.ShortcodeDomain())
			if _, _, err := p.load(ctx); err != nil {
				log.Errorf(ctx, "error loading emoji %s: %v", p.emoji.ShortcodeDomain(), err)
			}
		})
	}
	return emoji, err
}

// Placeholder returns a copy of internally stored processing placeholder,
// returning only the fields that may be known *before* completion, and as
// such all fields which are safe to concurrently read.
func (p *ProcessingEmoji) Placeholder() *gtsmodel.Emoji {
	emoji := new(gtsmodel.Emoji)
	emoji.ID = p.emoji.ID
	emoji.Shortcode = p.emoji.Shortcode
	emoji.Domain = p.emoji.Domain
	emoji.ImageRemoteURL = p.emoji.ImageRemoteURL
	emoji.ImageStaticRemoteURL = p.emoji.ImageStaticRemoteURL
	emoji.Disabled = p.emoji.Disabled
	emoji.VisibleInPicker = p.emoji.VisibleInPicker
	emoji.CategoryID = p.emoji.CategoryID
	emoji.UpdatedAt = p.emoji.UpdatedAt
	emoji.CreatedAt = p.emoji.CreatedAt

	var pathID string
	if p.newPathID != "" {
		pathID = p.newPathID
	} else {
		pathID = p.emoji.ID
	}

	// We specifically set placeholder URL values that allow an API user to fetch the appropriate
	// emoji, even if we don't know what filetype it is yet. (since we just parse the IDs from URL path).
	//
	// This way the API caller can (in the worst case that it hasn't loaded yet) attempt to fetch the emoji,
	// then block on ProcessingEmoji{}.Load() for the processing entry it gets from a call to the dereferencer.
	emoji.ImageURL = uris.URIForAttachment(p.instAccID, string(TypeEmoji), string(SizeOriginal), pathID, "loading")
	emoji.ImageStaticURL = uris.URIForAttachment(p.instAccID, string(TypeEmoji), string(SizeStatic), pathID, "png")
	return emoji
}

// load is the package private form of load() that is wrapped to catch context canceled.
func (p *ProcessingEmoji) load(ctx context.Context) (
	emoji *gtsmodel.Emoji,
	done bool,
	err error,
) {
	err = p.proc.Process(func() (err error) {
		if done = p.done; done {
			// Already proc'd.
			err = p.err
			return
		}

		defer func() {
			// This is only done when ctx NOT cancelled.
			if done = (err == nil || !errorsv2.IsV2(err,
				context.Canceled,
				context.DeadlineExceeded,
			)); done {
				// Processing finished,
				// whether error or not!

				// Anything from here, we
				// need to ensure happens
				// (i.e. no ctx canceled).
				ctx = context.WithoutCancel(ctx)

				// Store values.
				p.done = done
				p.err = err

				// On error or stub, ensure
				// no downloaded files remain.
				if err != nil {
					p.cleanup(ctx)
				}

				// Check the extracted error details on emoji for
				// stub type error. i.e. policy or media type issue.
				if isStubError(p.emoji.Error) {
					log.Warnf(ctx, "stubbed %s due to: %v", p.emoji.ImageRemoteURL, p.err)
					err = nil // don't return stub errors
				}

				// Update with latest details, whatever happened.
				e := p.mgr.state.DB.UpdateEmoji(ctx, p.emoji)
				if e != nil {
					log.Errorf(ctx, "error updating emoji in db: %v", e)
				}
			}
		}()

		// If existing error details exists, check if supports retry.
		if withDetails := errors.AsV2[*errWithDetails](p.err); //
		withDetails != nil && !withDetails.details.SupportsRetry() {
			err = p.err
			return
		}

		// Attempt to store media and calculate
		// full-size media attachment details.
		//
		// This will update p.emoji as it goes.
		err = p.store(ctx)
		return
	})

	// Return a copy of emoji.
	emoji = new(gtsmodel.Emoji)
	*emoji = *p.emoji
	return
}

// store calls the data function attached to p if it hasn't been called yet,
// and updates the underlying attachment fields as necessary. It will then stream
// bytes from p's reader directly into storage so that it can be retrieved later.
func (p *ProcessingEmoji) store(ctx context.Context) error {

	// Load media from data func.
	rc, err := p.dataFn(ctx)
	if err != nil {

		// If a network error, include these details.
		if details := extractNetworkErrorDetails(err); //
		details != 0 {
			err = withDetails(err, details)
		}

		return gtserror.Newf("error executing data function: %w", err)
	}

	var (
		// predfine temporary media
		// file path variables so we
		// can remove them on error.
		temppath   string
		staticpath string
	)

	defer func() {
		if err := remove(temppath, staticpath); err != nil {
			log.Errorf(ctx, "error(s) cleaning up files: %v", err)
		}
	}()

	// Drain reader to tmp file
	// (this reader handles close).
	temppath, err = drainToTmp(rc)
	if err != nil {
		return gtserror.Newf("error draining data to tmp: %w", err)
	}

	// Pass input file through ffprobe to
	// parse further metadata information.
	result, err := probe(ctx, temppath)
	if err != nil {
		return gtserror.Newf("ffprobe error: %w", err)
	}

	var ext string
	var fileType gtsmodel.FileType

	// Get abstract file type, mimetype and ext from ffprobe data.
	fileType, p.emoji.ImageContentType, ext = result.GetFileType()
	if fileType != gtsmodel.FileTypeImage {

		// NOTE: unlike regular media, unsupported file type is an error for emoji.
		err := gtserror.Newf("unsupported emoji filetype: %s (%s)", fileType, ext)
		return withDetails(err, codecDetails)
	}

	// Add file extension to path.
	newpath := temppath + "." + ext

	// Before ffmpeg processing, rename to set file ext.
	if err := os.Rename(temppath, newpath); err != nil {
		return gtserror.Newf("error renaming to %s - >%s: %w", temppath, newpath, err)
	}

	// Update path var
	// AFTER successful.
	temppath = newpath

	// Generate a static image from input emoji path.
	staticpath, err = ffmpegGenerateStatic(ctx, temppath)
	if err != nil {
		return gtserror.Newf("error generating emoji static: %w", err)
	}

	var pathID string
	if p.newPathID != "" {
		// This is a refreshed emoji with a new
		// path ID that this will be stored under.
		pathID = p.newPathID
	} else {
		// This is a new emoji, simply use provided ID.
		pathID = p.emoji.ID
	}

	// Calculate final emoji media file path.
	p.emoji.ImagePath = uris.StoragePathForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeOriginal),
		pathID,
		ext,
	)

	// Calculate final emoji static media file path.
	p.emoji.ImageStaticPath = uris.StoragePathForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeStatic),
		pathID,
		"png",
	)

	// Set the known emoji static content type.
	p.emoji.ImageStaticContentType = "image/png"

	// Copy temporary file into storage at path.
	filesz, err := p.mgr.state.Storage.PutFile(ctx,
		p.emoji.ImagePath,
		temppath,
		p.emoji.ImageContentType,
	)
	if err != nil {
		return gtserror.Newf("error writing emoji to storage: %w", err)
	}

	// Copy static emoji file into storage at path.
	staticsz, err := p.mgr.state.Storage.PutFile(ctx,
		p.emoji.ImageStaticPath,
		staticpath,
		p.emoji.ImageStaticContentType,
	)
	if err != nil {
		return gtserror.Newf("error writing static to storage: %w", err)
	}

	// Set final determined file sizes.
	p.emoji.ImageFileSize = int(filesz)
	p.emoji.ImageStaticFileSize = int(staticsz)

	// Generate an emoji media static URL.
	p.emoji.ImageURL = uris.URIForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeOriginal),
		pathID,
		ext,
	)

	// Generate an emoji image static URL.
	p.emoji.ImageStaticURL = uris.URIForAttachment(
		p.instAccID,
		string(TypeEmoji),
		string(SizeStatic),
		pathID,
		"png",
	)

	// Success! Unset previous
	// error details for emoji.
	p.emoji.Error = 0

	return nil
}

// cleanup will remove any traces of processing emoji from storage,
// and perform any other necessary cleanup steps after failure.
//
// details of any error are extracted and can be accessed via p.emoji.Error.
func (p *ProcessingEmoji) cleanup(ctx context.Context) {
	log.Debugf(ctx, "running cleanup of emoji %s", p.emoji.ID)

	if p.emoji.ImagePath != "" {
		// Ensure emoji file at path is deleted from storage.
		err := p.mgr.state.Storage.Delete(ctx, p.emoji.ImagePath)
		if err != nil && !storage.IsNotFound(err) {
			log.Errorf(ctx, "error deleting %s: %v", p.emoji.ImagePath, err)
		}
	}

	if p.emoji.ImageStaticPath != "" {
		// Ensure emoji static file at path is deleted from storage.
		err := p.mgr.state.Storage.Delete(ctx, p.emoji.ImageStaticPath)
		if err != nil && !storage.IsNotFound(err) {
			log.Errorf(ctx, "error deleting %s: %v", p.emoji.ImageStaticPath, err)
		}
	}

	// Unset fields.
	p.emoji.Stub()

	// Extract any error details for db.
	p.emoji.Error = toErrorDetails(p.err)
}
